/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v6
 * (C) Copyright IBM Corp. 2013, 2014. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * %W% %E%
 * 
 * Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * -Redistribution of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 * 
 * -Redistribution in binary form must reproduce the above copyright notice, 
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 * 
 * Neither the name of Oracle or the names of contributors may 
 * be used to endorse or promote products derived from this software without 
 * specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL 
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST 
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY 
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#if !defined(LINUX) && !defined(AIX)                                            //IBM-new_platforms
#include <procfs.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <unistd.h>
#include <errno.h>
#include <dlfcn.h>
#include <sys/time.h>

#include <netdb.h>
#include <netinet/in.h>
#ifdef __MVS__                                                                  //IBM-new_platforms
#include <arpa/inet.h>                                                          //IBM-new_platforms
#endif                                                                          //IBM-new_platforms
#include <sys/param.h>
#include <time.h>

#include "jni.h"
#include "hprof.h"

#ifdef __MVS__                                                                  //IBM-new_platforms
#include "atoe.h"                      /* for ASCII to EBCDIC conversion */     //IBM-new_platforms
#endif                                                                          //IBM-new_platforms
                                                                                //IBM-new_platforms
#define PATH_SEPARATOR          ":"
#define PATH_SEPARATOR_CHAR     ':'

int
md_getpid(void)
{
    static int pid = -1;
    
    if ( pid >= 0 ) {
	return pid;
    }
    pid = getpid();
    return pid;
}

void 
md_sleep(unsigned seconds)
{
    sleep(seconds);
}

void
md_init(void)
{
#if defined(LINUX) || defined(__MVS__) || defined(AIX)                          //IBM-new_platforms
    /* No Hi-Res timer option? */
#else
    if ( gdata->micro_state_accounting ) {
	char proc_ctl_fn[48];
	int  procfd;

	/* Turn on micro state accounting, once per process */
	(void)md_snprintf(proc_ctl_fn, sizeof(proc_ctl_fn),
		"/proc/%d/ctl", md_getpid());
     
	procfd = open(proc_ctl_fn, O_WRONLY);
	if (procfd >= 0) {
	    long ctl_op[2];
	    
	    ctl_op[0] = PCSET;
	    ctl_op[1] = PR_MSACCT;
	    (void)write(procfd, ctl_op, sizeof(ctl_op));
	    (void)close(procfd);
	}
    }
#endif
}

int
md_connect(char *hostname, unsigned short port)
{
    struct hostent *hentry;
    struct sockaddr_in s;
    int fd;

    /* create a socket */
    fd = socket(AF_INET, SOCK_STREAM, 0);

    /* find remote host's addr from name */
    if ((hentry = gethostbyname(hostname)) == NULL) {
	return -1;
    }
    (void)memset((char *)&s, 0, sizeof(s));
    /* set remote host's addr; its already in network byte order */
    (void)memcpy(&s.sin_addr.s_addr, *(hentry->h_addr_list),
	   (int)sizeof(s.sin_addr.s_addr));
    /* set remote host's port */
    s.sin_port = htons(port);
    s.sin_family = AF_INET;

    /* now try connecting */
    if (-1 == connect(fd, (struct sockaddr*)&s, sizeof(s))) {
	return 0;
    }
    return fd;
}

int
md_recv(int f, char *buf, int len, int option)
{
    return recv(f, buf, len, option);
}

int
md_shutdown(int filedes, int option)
{
    return shutdown(filedes, option);
}

int
md_open(const char *filename)
{
    return open(filename, O_RDONLY);
}

int
md_open_binary(const char *filename)
{
    return md_open(filename);
}

int
md_creat(const char *filename)
{
    return open(filename, O_WRONLY | O_CREAT | O_TRUNC,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
}

int
md_creat_binary(const char *filename)
{
    return md_creat(filename);
}

jlong
md_seek(int filedes, jlong cur)
{
    jlong new_pos;

    if ( cur == (jlong)-1 ) {
        new_pos = lseek(filedes, 0, SEEK_END);
    } else {
        new_pos = lseek(filedes, cur, SEEK_SET);
    }
    return new_pos;
}

void
md_close(int filedes)
{
    (void)close(filedes);
}

int 
md_send(int s, const char *msg, int len, int flags)
{
    int res;
    
    do {
        res = send(s, msg, len, flags);
    } while ((res < 0) && (errno == EINTR));
    
    return res;
}

/*                                                                              //IBM-new_platforms
 * In zOS, while writing the data, we have to ensure that the                   //IBM-new_platforms
 * data is written in EBCDIC format. If the data is in ASCII                    //IBM-new_platforms
 * format, it has to be converted to EBCDIC.                                    //IBM-new_platforms
 */                                                                             //IBM-new_platforms
int 
md_write(int filedes, const void *buf, int nbyte)
{
    int res;
    
#if defined(__MVS__)                                                            //IBM-new_platforms
    void *write_buf;					                    //IBM-new_platforms
    size_t write_nbyte;					                 //IBM-new_platforms
                                                                                //IBM-new_platforms
    if ('a' == gdata->output_format) {                                          //IBM-new_platforms
        write_buf = a2e(buf, nbyte);		                            //IBM-new_platforms
        write_nbyte = (size_t)nbyte;                                            //IBM-new_platforms
        do {                                                                    //IBM-new_platforms
            res = (int)write(filedes, write_buf, write_nbyte);                  //IBM-new_platforms
        } while ((res < 0) && (errno == EINTR));                                //IBM-new_platforms
        free(write_buf);                                                        //IBM-new_platforms
    } else {                                                                    //IBM-new_platforms
#endif                                                                          //IBM-new_platforms
        do {                                                                    //IBM-new_platforms
            res = (int)write(filedes, buf, nbyte);                                  //IBM-new_platforms
        } while ((res < 0) && (errno == EINTR));                                //IBM-new_platforms

#if defined(__MVS__)                                                            //IBM-new_platforms
    }                                                                           //IBM-new_platforms
#endif                                                                          //IBM-new_platforms
    return res;
}

/*                                                                              //IBM-new_platforms
 * In zOS, while reading the data, the data will be available                   //IBM-new_platforms
 * in ASCII format. If the output is expected in ASCII format,                  //IBM-new_platforms
 * the EBCDIC data has to be converted to ASCII.                                //IBM-new_platforms
 */                                                                             //IBM-new_platforms
int
md_read(int filedes, void *buf, int nbyte)
{
    int res;
    
    do {
        res = (int)read(filedes, buf, nbyte);                                   //IBM-new_platforms
    } while ((res < 0) && (errno == EINTR));

#if defined(__MVS__)                                                            //IBM-new_platforms
    void *read_buf;						             //IBM-new_platforms
                                                                                //IBM-new_platforms
    if ('a' == gdata->output_format) {			                  //IBM-new_platforms
        read_buf = e2a(buf, (size_t)(nbyte - 1));                               //IBM-new_platforms
        memcpy(buf, read_buf, (size_t)nbyte);                                   //IBM-new_platforms
        free(read_buf);                                                         //IBM-new_platforms
    }							                   //IBM-new_platforms
#endif                                                                          //IBM-new_platforms
                                                                                //IBM-new_platforms
    return res;
}

/* Time of day in milli-seconds */
static jlong
md_timeofday(void)
{
    struct timeval tv;
 
    if ( gettimeofday(&tv, (void *)0) != 0 ) {
	return (jlong)0; /* EOVERFLOW ? */
    }
    /*LINTED*/
    return ((jlong)tv.tv_sec * (jlong)1000) + (jlong)(tv.tv_usec / 1000);
}

/* Hi-res timer in micro-seconds */
jlong 
md_get_microsecs(void)
{
#if defined(LINUX) || defined(__MVS__) || defined(AIX)                          //IBM-new_platforms
    return (jlong)(md_timeofday() * (jlong)1000); /* Milli to micro */
#else
    return (jlong)(gethrtime()/(hrtime_t)1000); /* Nano seconds to micro seconds */
#endif
}

/* Time of day in milli-seconds */
jlong 
md_get_timemillis(void)
{
    return md_timeofday();
}

/* Current CPU hi-res CPU time used */
jlong
md_get_thread_cpu_timemillis(void)
{
#if defined(LINUX) || defined(__MVS__) || defined(AIX)                          //IBM-new_platforms
    return md_timeofday();
#else
    return (jlong)(gethrvtime()/1000); /* Nano seconds to milli seconds */
#endif
}

/*                                                                              //IBM-new_platforms
 * In AIX and zOS, a different method have to be used to identify               //IBM-new_platforms
 * the path to the filename. First identify the path to "libhprof.so" and       //IBM-new_platforms
 * then identify the path to filename relative to the path of "libhprof.so".    //IBM-new_platforms
 */                                                                             //IBM-new_platforms
void 
md_get_prelude_path(char *path, int path_len, char *filename)
{
#if defined(AIX) || defined(__MVS__)                                            //IBM-new_platforms
{                                                                               //IBM-new_platforms
    struct stat sbuf;                                                           //IBM-new_platforms
    char tmp_path[MAXPATHLEN];                                                  //IBM-new_platforms
    static char *pathspec = NULL;                                               //IBM-new_platforms
    char *saveP = NULL;                                                         //IBM-new_platforms
    char pathbuf[MAXPATHLEN];                                                   //IBM-new_platforms
    char *subpath;                                                              //IBM-new_platforms
                                                                                //IBM-new_platforms
    if (!pathspec) {                                                            //IBM-new_platforms
        pathspec = getenv("LIBPATH");                                           //IBM-new_platforms
    }                                                                           //IBM-new_platforms
                                                                                //IBM-new_platforms
    if (pathspec == (char *)0 || *pathspec == '\0') {                           //IBM-new_platforms
        *path = '\0';                                                           //IBM-new_platforms
        return;                                                                 //IBM-new_platforms
    }                                                                           //IBM-new_platforms
                                                                                //IBM-new_platforms
    strcpy(pathbuf, pathspec);                                                  //IBM-new_platforms
    subpath = strtok_r(pathbuf,":", &saveP);                                    //IBM-new_platforms
                                                                                //IBM-new_platforms
    while (subpath != (char *)0)                                                //IBM-new_platforms
    {                                                                           //IBM-new_platforms
        strcpy(tmp_path,subpath);                                               //IBM-new_platforms
        strcat(tmp_path,"/");                                                   //IBM-new_platforms
        strcat(tmp_path,"libhprof.so");  /*AIX_FIXME - can we do better?*/      //IBM-new_platforms
                                                                                //IBM-new_platforms
        if (stat(tmp_path, &sbuf) == 0) {                                       //IBM-new_platforms
            snprintf(path, path_len, "%s/../../lib/%s", subpath, filename);     //IBM-new_platforms
            return;                                                             //IBM-new_platforms
        }                                                                       //IBM-new_platforms
        subpath = strtok_r(NULL,":", &saveP);                                   //IBM-new_platforms
    }                                                                           //IBM-new_platforms
    *path = '\0';                                                               //IBM-new_platforms
    return;                                                                     //IBM-new_platforms
}                                                                               //IBM-new_platforms
                                                                                //IBM-new_platforms
#else                                                                           //IBM-new_platforms
                                                                                //IBM-new_platforms
    void *addr;
    char libdir[FILENAME_MAX+1];
    Dl_info dlinfo;

    libdir[0] = 0;
#if defined(LINUX)                                                              //IBM-new_platforms
    addr = (void*)&Agent_OnLoad;
#else
    /* Just using &Agent_OnLoad will get the first external symbol with
     *   this name in the first .so, which may not be libhprof.so.
     *   On Solaris we can actually ask for the address of our Agent_OnLoad.
     */
    addr = dlsym(RTLD_SELF, "Agent_OnLoad");
    /* Just in case the above didn't work (missing linker patch?). */
    if ( addr == NULL ) {
	addr = (void*)&Agent_OnLoad;
    }
#endif

    /* Use dladdr() to get the full path to libhprof.so, which we use to find
     *  the prelude file.
     */
    dlinfo.dli_fname = NULL;
    (void)dladdr(addr, &dlinfo);
    if ( dlinfo.dli_fname != NULL ) {
	char * lastSlash;

	/* Full path to library name, need to move up one directory to 'lib' */
	(void)strcpy(libdir, (char *)dlinfo.dli_fname);
	lastSlash = strrchr(libdir, '/');
	if ( lastSlash != NULL ) {
	    *lastSlash = '\0';
	}
	lastSlash = strrchr(libdir, '/');
	if ( lastSlash != NULL ) {
	    *lastSlash = '\0';
	}
    }
    (void)snprintf(path, path_len, "%s/%s", libdir, filename);
#endif                                                                          //IBM-new_platforms
}


int     
md_vsnprintf(char *s, int n, const char *format, va_list ap)
{
    return vsnprintf(s, n, format, ap);
}

int     
md_snprintf(char *s, int n, const char *format, ...)
{
    int ret;
    va_list ap;

    va_start(ap, format);
    ret = md_vsnprintf(s, n, format, ap);
    va_end(ap);
    return ret;
}

void
md_system_error(char *buf, int len)
{
    char *p;
    
    buf[0] = 0;
    p = strerror(errno);
    if ( p != NULL ) {
	(void)strcpy(buf, p);
    }
}

unsigned
md_htons(unsigned short s)
{
    return htons(s);
}

unsigned
md_htonl(unsigned l)
{
    return htonl(l);
}

unsigned	
md_ntohs(unsigned short s)
{
    return ntohs(s);
}

unsigned
md_ntohl(unsigned l)
{
    return ntohl(l);
}

/*
 * splits a path, based on its separator, the number of
 * elements is returned back in n. 
 * It is the callers responsibility to:
 *   a> check the value of n, and n may be 0
 *   b> ignore any empty path elements
 *   c> free up the data.
 */
static char** split_path(const char* path, int* n) {
    char* inpath;
    char** opath;
    char* p;
    int count = 1;
    int i;
    
    *n = 0;
    if (path == NULL || strlen(path) == 0) {
        return NULL;
    }
    inpath = strdup(path);
    if (inpath == NULL) {
        return NULL;
    }
    p = strchr(inpath, PATH_SEPARATOR_CHAR);
    // get a count of elements to allocate memory
    while (p != NULL) {
        count++;
        p++;
        p = strchr(p, PATH_SEPARATOR_CHAR);
    }
    opath = (char**) calloc(count, sizeof(char*));
    if (opath == NULL) {
        return NULL;
    }

    // do the actual splitting
    p = inpath;
    for (i = 0 ; i < count ; i++) {
        size_t len = strcspn(p, PATH_SEPARATOR);
	// allocate the string and add terminator storage
        char* s  = (char*)malloc((len + 1)*sizeof(char));
        if (s == NULL) {
            return NULL;
	}
        strncpy(s, p, len);
	s[len] = '\0';
        opath[i] = s;
        p += len + 1;
    }
    free(inpath);
    *n = count;
    return opath;
}

/* Create the actual fill filename for a dynamic library.  */
void
md_build_library_name(char *holder, int holderlen, char *pname, char *fname)
{
    int n;
    int i;
    char** pelements;
    struct stat statbuf;
    const int pnamelen = pname ? strlen(pname) : 0;

    /* Quietly truncate on buffer overflow.  Should be an error. */
    if (pnamelen + (int)strlen(fname) + 10 > holderlen) {
        *holder = '\0';
        return;
    }

    /* Construct path to library */
    if (pnamelen == 0) {
        (void)snprintf(holder, holderlen, "lib%s.so", fname);
    } else if (strchr(pname, PATH_SEPARATOR_CHAR) != NULL) {
        pelements = split_path(pname, &n);
        for (i = 0 ; i < n ; i++) {
            // really shouldn't be NULL but what the heck, check can't hurt
	    if (pelements[i] == NULL || strlen(pelements[i]) == 0) {
	       continue; // skip the empty path values
	    }
	    snprintf(holder, holderlen, "%s/lib%s.so", pelements[i], fname);
	    if (stat(holder, &statbuf) == 0) {
	       break;
	    }
        }
        // release the storage
        for (i = 0 ; i < n ; i++) {
            if (pelements[i] != NULL) {
               free(pelements[i]);
            }
        }
        if (pelements != NULL) {
            free(pelements);
        }
    } else {
        (void)snprintf(holder, holderlen, "%s/lib%s.so", pname, fname);
    }
}

/* Load this library (return NULL on error, and error message in err_buf) */
void *
md_load_library(const char *name, char *err_buf, int err_buflen)
{
    void * result;
    
    result = dlopen(name, RTLD_LAZY);
    if (result == NULL) {
	(void)strncpy(err_buf, dlerror(), err_buflen-2);
	err_buf[err_buflen-1] = '\0';
    }
    return result;
}

/* Unload this library */
void 
md_unload_library(void *handle)
{
    (void)dlclose(handle);
}

/* Find an entry point inside this library (return NULL if not found) */
void * 
md_find_library_entry(void *handle, const char *name)
{
    void * sym;
    
    sym =  dlsym(handle, name);
    return sym;
}

//IBM-new_platforms
